from typing import Protocol

from .song import PlaybackState


class Player(Protocol):
	async def on_play_pause(self) -> PlaybackState:
		...

	async def on_play(self) -> PlaybackState:
		...

	async def on_pause(self) -> PlaybackState:
		...

	async def on_stop(self) -> PlaybackState:
		...

	async def on_next(self) -> None:
		...

	async def on_prev(self) -> None:
		...

	async def on_seek(self, position: float) -> None:
		...



from enum import StrEnum
from pathlib import Path
from typing import Protocol
from uuid import UUID

from attrs import define, field


class PlaybackState(StrEnum):
	play = "play"
	pause = "pause"
	stop = "stop"


@define
class Song:
	state: PlaybackState
	queue_index: int
	queue_length: int
	file: Path
	musicbrainz_trackid: UUID | None
	musicbrainz_releasetrackid: UUID | None
	title: str | None
	artist: str | None
	composer: str | None
	album: str | None
	album_artist: str | None
	track: int | None
	disc: int | None
	genre: str | None
	duration: float
	elapsed: float
	art: bytes | None = field(repr=lambda a: "<has art>" if a else "<no art>")


class SongListener(Protocol):
	def update(self, song: Song | None) -> None:
		...



# -*- coding: utf-8 -*-
# Copied from https://github.com/alberthier/corefoundationasyncio/blob/5061b9b7daa8bcd40d54d58432d84dcc0a339ca6/corefoundationasyncio.py
# This module is copied, rather than simply installed, because the PyPI version of this module depends on *all* of PyObjC even though it only needs pyobjc-framework-Cocoa.
# There's an open PR to fix this: httpsV//github.com/alberthier/corefoundationasyncio/pull/3


import asyncio
import sys
import threading

from CoreFoundation import (
    CFRunLoopGetCurrent,
    CFRunLoopTimerCreate, CFRunLoopAddTimer, CFRunLoopRemoveTimer, CFAbsoluteTimeGetCurrent,
    CFFileDescriptorCreate, CFFileDescriptorIsValid, CFFileDescriptorEnableCallBacks, CFFileDescriptorDisableCallBacks,
    CFFileDescriptorCreateRunLoopSource, CFRunLoopAddSource, CFRunLoopRemoveSource,
    kCFAllocatorDefault, kCFRunLoopDefaultMode, kCFRunLoopCommonModes,
    kCFFileDescriptorReadCallBack, kCFFileDescriptorWriteCallBack
)
import PyObjCTools.AppHelper

class _TimerHandle(asyncio.TimerHandle):

    def __init__(self, when, callback, args, loop, context):
        super().__init__(when, callback, args, loop, context)
        self.cf_runloop_timer = None


class _FDEntry:

    def __init__(self):
        self.cf_fd = None
        self.cf_source = None
        self.callbacks = {}


class CoreFoundationEventLoop(asyncio.SelectorEventLoop):
    """
    Event loop based on CoreFoundation's CFRunLoop.
    This allows integration of Cocoa GUI apps with asyncio
    """

    def __init__(self, console_app = False, *eventloop_args):
        self._console_app = console_app
        self._eventloop_args = eventloop_args
        self._registered_fds = {}
        self._runloop = CFRunLoopGetCurrent()
        super().__init__()

    # Running and stopping the event loop.

    def run_forever(self):
        """Run until stop() is called."""
        self._check_closed()
        if self.is_running():
            raise RuntimeError('This event loop is already running')
        if asyncio._get_running_loop() is not None:
            raise RuntimeError('Cannot run the event loop while another loop is running')
        self._set_coroutine_origin_tracking(self._debug)
        self._thread_id = threading.get_ident()

        old_agen_hooks = sys.get_asyncgen_hooks()
        sys.set_asyncgen_hooks(firstiter=self._asyncgen_firstiter_hook,
                               finalizer=self._asyncgen_finalizer_hook)
        try:
            asyncio._set_running_loop(self)
            if self._console_app:
                PyObjCTools.AppHelper.runConsoleEventLoop(*self._eventloop_args)
            else:
                PyObjCTools.AppHelper.runEventLoop(*self._eventloop_args)
        finally:
            self._thread_id = None
            asyncio._set_running_loop(None)
            self._set_coroutine_origin_tracking(False)
            sys.set_asyncgen_hooks(*old_agen_hooks)

    def _process_events(self, event_list):
        raise NotImplementedError("Not available in this implementation")

    def _run_once(self):
        raise NotImplementedError("Not available in this implementation")

    def stop(self):
        PyObjCTools.AppHelper.stopEventLoop()

    # Methods scheduling callbacks.  All these return Handles.

    def call_at(self, when, callback, *args, context=None):
        self._check_closed()
        if self._debug:
            self._check_thread()
            self._check_callback(callback, 'call_at')
        timerHandle = _TimerHandle(when, callback, args, self, context)
        if timerHandle._source_traceback:
            del timerHandle._source_traceback[-1]
        self._add_callback(timerHandle)
        return timerHandle

    def _call_soon(self, callback, args, context):
        handle = asyncio.Handle(callback, args, self, context)
        self._add_callback(handle)
        return handle

    def _add_callback(self, handle):
        assert isinstance(handle, asyncio.Handle), 'A Handle is required here'
        is_timer = isinstance(handle, _TimerHandle)
        if handle.cancelled():
            return
        def ontimeout(cf_timer, info):
            if not handle.cancelled():
                handle._run()
        when = handle.when() if is_timer else self.time()
        cf_timer = CFRunLoopTimerCreate(kCFAllocatorDefault, when, 0, 0, 0, ontimeout, None)
        CFRunLoopAddTimer(self._runloop, cf_timer, kCFRunLoopCommonModes)
        if is_timer:
            handle.cf_runloop_timer = cf_timer
            handle._scheduled = True

    def _timer_handle_cancelled(self, handle):
        CFRunLoopRemoveTimer(self._runloop, handle.cf_runloop_timer, kCFRunLoopCommonModes)

    def time(self):
        return CFAbsoluteTimeGetCurrent()

    # Ready-based callback registration methods.
    # The add_*() methods return None.
    # The remove_*() methods return True if something was removed,
    # False if there was nothing to delete.

    def _register_fd(self, fd, event, callback, args):
        entry = self._registered_fds.get(fd)
        if entry is not None:
            CFFileDescriptorEnableCallBacks(entry.cf_fd, event)
            entry.callbacks[event] = (callback, args)
        else:
            def fd_callback(file_desc, callback_types, entry):
                if callback_types & kCFFileDescriptorReadCallBack:
                    (callback, args) = entry.callbacks[kCFFileDescriptorReadCallBack]
                    callback(*args)
                if callback_types & kCFFileDescriptorWriteCallBack:
                    (callback, args) = entry.callbacks[kCFFileDescriptorWriteCallBack]
                    callback(*args)
                if CFFileDescriptorIsValid(file_desc):
                    CFFileDescriptorEnableCallBacks(file_desc, callback_types)

            entry = _FDEntry()
            entry.cf_fd = CFFileDescriptorCreate(kCFAllocatorDefault, fd, 0, fd_callback, entry)
            CFFileDescriptorEnableCallBacks(entry.cf_fd, event)
            entry.callbacks[event] = (callback, args)
            entry.cf_source = CFFileDescriptorCreateRunLoopSource(kCFAllocatorDefault, entry.cf_fd, 0)
            self._registered_fds[fd] = entry
            CFRunLoopAddSource(self._runloop, entry.cf_source, kCFRunLoopDefaultMode)

    def _unregister_fd(self, fd, event):
        entry = self._registered_fds.pop(fd, None)
        if entry is None:
            return False
        cb = entry.callbacks.pop(event, None)
        if cb is None:
            return False
        if len(entry.callbacks) != 0:
            CFFileDescriptorDisableCallBacks(entry.cf_fd, event)
        else:
            CFRunLoopRemoveSource(self._runloop, entry.cf_source, kCFRunLoopDefaultMode)
        return True

    def _add_reader(self, fd, callback, *args):
        self._check_closed()
        self._register_fd(fd, kCFFileDescriptorReadCallBack, callback, args)

    def _remove_reader(self, fd):
        if self.is_closed():
            return False
        return self._unregister_fd(fd, kCFFileDescriptorReadCallBack)

    def _add_writer(self, fd, callback, *args):
        self._check_closed()
        self._register_fd(fd, kCFFileDescriptorWriteCallBack, callback, args)

    def _remove_writer(self, fd):
        if self.is_closed():
            return False
        return self._unregister_fd(fd, kCFFileDescriptorWriteCallBack)



from hashlib import blake2b
from pathlib import Path
from typing import Final
from uuid import UUID

from ..song import Song

# The maximum size for a BLAKE2b "person" value is sixteen bytes, so we need to be concise.
HASH_PERSON_PREFIX: Final = b"mnp.mac."
TRACKID_HASH_PERSON: Final = HASH_PERSON_PREFIX + b"mb_tid"
RELEASETRACKID_HASH_PERSON: Final = HASH_PERSON_PREFIX + b"mb_rtid"
FILE_HASH_PERSON: Final = HASH_PERSON_PREFIX + b"f"

PERSISTENT_ID_BITS: Final = 64
PERSISTENT_ID_BYTES: Final = PERSISTENT_ID_BITS // 8


def digest_trackid(trackid: UUID) -> bytes:
	return blake2b(
		trackid.bytes, digest_size=PERSISTENT_ID_BYTES, person=TRACKID_HASH_PERSON
	).digest()


def digest_releasetrackid(trackid: UUID) -> bytes:
	return blake2b(
		trackid.bytes,
		digest_size=PERSISTENT_ID_BYTES,
		person=RELEASETRACKID_HASH_PERSON,
	).digest()


def digest_file_uri(file: Path) -> bytes:
	return blake2b(
		bytes(file), digest_size=PERSISTENT_ID_BYTES, person=FILE_HASH_PERSON
	).digest()


# The MPMediaItemPropertyPersistentID is only 64 bits, while a UUID is 128
# bits and not all tracks will even have their MusicBrainz track ID included.
# To work around this, we compute a BLAKE2 hash from the UUID, or failing
# that from the file URI. BLAKE2 can be customised to different digest sizes,
# making it perfect for this problem.
def song_to_persistent_id(song: Song) -> int:
	if song.musicbrainz_trackid:
		hashed_id = digest_trackid(song.musicbrainz_trackid)
	elif song.musicbrainz_releasetrackid:
		hashed_id = digest_releasetrackid(song.musicbrainz_releasetrackid)
	else:
		hashed_id = digest_file_uri(song.file)
	return int.from_bytes(hashed_id)



import asyncio
from collections.abc import Coroutine
from contextvars import Context
from typing import Optional

__all__ = ("run_background_task",)

background_tasks: set[asyncio.Task[None]] = set()


def run_background_task(
	coro: Coroutine[None, None, None],
	*,
	name: Optional[str] = None,
	context: Optional[Context] = None,
) -> None:
	task = asyncio.create_task(coro, name=name, context=context)
	background_tasks.add(task)
	task.add_done_callback(background_tasks.discard)



from typing import Literal, NotRequired, Protocol, TypedDict


class MpdStateHandler(Protocol):
	async def readpicture(self, file: str) -> bytes | None:
		...

	async def refresh(self) -> None:
		...


BooleanFlag = Literal["0", "1"]

OneshotFlag = Literal[BooleanFlag, "oneshot"]


# This is not the complete status response from MPD, just the parts of it mpd-now-playable uses.
class StatusResponse(TypedDict):
	state: Literal["play", "stop", "pause"]

	# The total duration and elapsed playback of the current song, measured in seconds. Fractional seconds are allowed.
	duration: str
	elapsed: str

	# The volume value ranges from 0-100. It may be omitted from
	# the response entirely if MPD has no volume mixer configured.
	volume: NotRequired[str]

	# Various toggle-able music playback settings, which can be addressed and modified by Now Playing.
	repeat: BooleanFlag
	random: BooleanFlag
	single: OneshotFlag
	consume: OneshotFlag

	# Partitions essentially let one MPD server act as multiple music players.
	# For most folks, this will just be "default", but mpd-now-playable will
	# eventually support addressing specific partitions. Eventually.
	partition: str

	# The total number of items in the play queue, which is called the "playlist" throughout the MPD protocol for legacy reasons.
	playlistlength: str

	# The format of decoded audio MPD is producing, expressed as a string in the form "samplerate:bits:channels".
	audio: str


# All of these are metadata tags read from your music, and are strictly
# optional. mpd-now-playable will work better if your music is properly
# tagged, since then it can pass more information on to Now Playing, but it
# should work fine with completely untagged music too.
class CurrentSongTags(TypedDict, total=False):
	artist: str
	albumartist: str
	artistsort: str
	albumartistsort: str
	title: str
	album: str
	track: str
	date: str
	originaldate: str
	composer: str
	disc: str
	label: str
	genre: str
	musicbrainz_albumid: str
	musicbrainz_albumartistid: str
	musicbrainz_releasetrackid: str
	musicbrainz_artistid: str
	musicbrainz_trackid: str


class CurrentSongResponse(CurrentSongTags):
	# The name of the music file currently being played, as MPD understands
	# it. For locally stored music files, this'll just be a simple file path
	# relative to your music directory.
	file: str
	# The index of the song in the play queue. Will change if you shuffle or
	# otherwise reorder the playlist.
	pos: str


ReadPictureResponse = TypedDict("ReadPictureResponse", {"binary": bytes})



import asyncio
from pathlib import Path
from uuid import UUID

from mpd.asyncio import MPDClient
from mpd.base import CommandError

from ..player import Player
from ..song import PlaybackState, Song, SongListener
from ..type_tools import convert_if_exists
from .artwork_cache import MpdArtworkCache
from .types import CurrentSongResponse, StatusResponse


def mpd_current_to_song(
	status: StatusResponse, current: CurrentSongResponse, art: bytes | None
) -> Song:
	return Song(
		state=PlaybackState(status["state"]),
		queue_index=int(current["pos"]),
		queue_length=int(status["playlistlength"]),
		file=Path(current["file"]),
		musicbrainz_trackid=convert_if_exists(current.get("musicbrainz_trackid"), UUID),
		musicbrainz_releasetrackid=convert_if_exists(
			current.get("musicbrainz_releasetrackid"), UUID
		),
		title=current.get("title"),
		artist=current.get("artist"),
		album=current.get("album"),
		album_artist=current.get("albumartist"),
		composer=current.get("composer"),
		genre=current.get("genre"),
		track=convert_if_exists(current.get("track"), int),
		disc=convert_if_exists(current.get("disc"), int),
		duration=float(status["duration"]),
		elapsed=float(status["elapsed"]),
		art=art,
	)


class MpdStateListener(Player):
	client: MPDClient
	listener: SongListener
	art_cache: MpdArtworkCache
	idle_count = 0

	def __init__(self, cache: str | None = None) -> None:
		self.client = MPDClient()
		self.art_cache = (
			MpdArtworkCache(self, cache) if cache else MpdArtworkCache(self)
		)

	async def start(
		self, host: str = "localhost", port: int = 6600, password: str | None = None
	) -> None:
		print(f"Connecting to MPD server {host}:{port}...")
		await self.client.connect(host, port)
		if password is not None:
			print("Authorising to MPD with your password...")
			await self.client.password(password)
		print(f"Connected to MPD v{self.client.mpd_version}")

	async def refresh(self) -> None:
		await self.update_listener(self.listener)

	async def loop(self, listener: SongListener) -> None:
		self.listener = listener
		# notify our listener of the initial state MPD is in when this script loads up.
		await self.update_listener(listener)
		# then wait for stuff to change in MPD. :)
		async for _ in self.client.idle():
			self.idle_count += 1
			await self.update_listener(listener)

	async def update_listener(self, listener: SongListener) -> None:
		# If any async calls in here take long enough that we got another MPD idle event, we want to bail out of this older update.
		starting_idle_count = self.idle_count
		status, current = await asyncio.gather(
			self.client.status(), self.client.currentsong()
		)

		if starting_idle_count != self.idle_count:
			return

		if status["state"] == "stop":
			print("Nothing playing")
			listener.update(None)
			return

		art = await self.art_cache.get_cached_artwork(current)
		if starting_idle_count != self.idle_count:
			return

		song = mpd_current_to_song(status, current, art)
		print(song)
		listener.update(song)

	async def readpicture(self, file: str) -> bytes | None:
		try:
			readpic = await self.client.readpicture(file)
			return readpic.get("binary")
		except CommandError:
			return None

	async def on_play_pause(self) -> PlaybackState:
		# python-mpd2 has direct support for toggling the play/pause state by
		# calling MPDClient.pause(None), but it doesn't tell you the final
		# state, and we also want to support playing from stopped, so we need
		# to handle this ourselves.
		status = await self.client.status()
		return await {
			"play": self.on_pause,
			"pause": self.on_play,
			"stop": self.on_play,
		}[status["state"]]()

	async def on_play(self) -> PlaybackState:
		await self.client.play()
		return PlaybackState.play

	async def on_pause(self) -> PlaybackState:
		await self.client.pause(1)
		return PlaybackState.pause

	async def on_stop(self) -> PlaybackState:
		await self.client.stop()
		return PlaybackState.stop

	async def on_next(self) -> None:
		await self.client.next()

	async def on_prev(self) -> None:
		await self.client.previous()

	async def on_seek(self, position: float) -> None:
		await self.client.seekcur(position)



from typing import Callable, TypeVar

__all__ = ("convert_if_exists",)

T = TypeVar("T")


def convert_if_exists(value: str | None, converter: Callable[[str], T]) -> T | None:
	if value is None:
		return None
	return converter(value)



import asyncio
from os import environ

from corefoundationasyncio import CoreFoundationEventLoop

from .cocoa.now_playing import CocoaNowPlaying
from .mpd.listener import MpdStateListener


async def listen() -> None:
	port = int(environ.get("MPD_PORT", "6600"))
	host = environ.get("MPD_HOST", "localhost")
	password = environ.get("MPD_PASSWORD")
	cache = environ.get("MPD_NOW_PLAYABLE_CACHE")
	if password is None and "@" in host:
		password, host = host.split("@", maxsplit=1)

	listener = MpdStateListener(cache)
	now_playing = CocoaNowPlaying(listener)
	await listener.start(host=host, port=port, password=password)
	await listener.loop(now_playing)


def make_loop() -> CoreFoundationEventLoop:
	return CoreFoundationEventLoop(console_app=True)


def main() -> None:
	asyncio.run(listen(), loop_factory=make_loop, debug=True)


if __name__ == "__main__":
	main()



from collections.abc import Callable, Coroutine
from pathlib import Path

from AppKit import NSCompositingOperationCopy, NSImage, NSMakeRect
from Foundation import CGSize, NSMutableDictionary
from MediaPlayer import (
	MPChangePlaybackPositionCommandEvent,
	MPMediaItemArtwork,
	MPMediaItemPropertyAlbumTitle,
	MPMediaItemPropertyAlbumTrackNumber,
	MPMediaItemPropertyArtist,
	MPMediaItemPropertyArtwork,
	MPMediaItemPropertyComposer,
	MPMediaItemPropertyDiscNumber,
	MPMediaItemPropertyGenre,
	MPMediaItemPropertyPersistentID,
	MPMediaItemPropertyPlaybackDuration,
	MPMediaItemPropertyTitle,
	MPMusicPlaybackState,
	MPMusicPlaybackStatePaused,
	MPMusicPlaybackStatePlaying,
	MPMusicPlaybackStateStopped,
	MPNowPlayingInfoCenter,
	MPNowPlayingInfoMediaTypeAudio,
	MPNowPlayingInfoMediaTypeNone,
	MPNowPlayingInfoPropertyElapsedPlaybackTime,
	MPNowPlayingInfoPropertyExternalContentIdentifier,
	MPNowPlayingInfoPropertyMediaType,
	MPNowPlayingInfoPropertyPlaybackQueueCount,
	MPNowPlayingInfoPropertyPlaybackQueueIndex,
	MPNowPlayingInfoPropertyPlaybackRate,
	MPRemoteCommandCenter,
	MPRemoteCommandEvent,
	MPRemoteCommandHandlerStatus,
	MPRemoteCommandHandlerStatusSuccess,
)

from ..async_tools import run_background_task
from ..player import Player
from ..song import PlaybackState, Song
from .persistent_id import song_to_persistent_id


def logo_to_ns_image() -> NSImage:
	return NSImage.alloc().initByReferencingFile_(
		str(Path(__file__).parent.parent / "mpd/logo.svg")
	)


def data_to_ns_image(data: bytes) -> NSImage:
	return NSImage.alloc().initWithData_(data)


def ns_image_to_media_item_artwork(img: NSImage) -> MPMediaItemArtwork:
	def resize(size: CGSize) -> NSImage:
		new = NSImage.alloc().initWithSize_(size)
		new.lockFocus()
		img.drawInRect_fromRect_operation_fraction_(
			NSMakeRect(0, 0, size.width, size.height),
			NSMakeRect(0, 0, img.size().width, img.size().height),
			NSCompositingOperationCopy,
			1.0,
		)
		new.unlockFocus()
		return new

	return MPMediaItemArtwork.alloc().initWithBoundsSize_requestHandler_(
		img.size(), resize
	)


def playback_state_to_cocoa(state: PlaybackState) -> MPMusicPlaybackState:
	mapping: dict[PlaybackState, MPMusicPlaybackState] = {
		PlaybackState.play: MPMusicPlaybackStatePlaying,
		PlaybackState.pause: MPMusicPlaybackStatePaused,
		PlaybackState.stop: MPMusicPlaybackStateStopped,
	}
	return mapping[state]


def song_to_media_item(song: Song) -> NSMutableDictionary:
	nowplaying_info = nothing_to_media_item()
	nowplaying_info[MPNowPlayingInfoPropertyMediaType] = MPNowPlayingInfoMediaTypeAudio
	nowplaying_info[MPNowPlayingInfoPropertyElapsedPlaybackTime] = song.elapsed
	nowplaying_info[MPNowPlayingInfoPropertyExternalContentIdentifier] = str(song.file)
	nowplaying_info[MPNowPlayingInfoPropertyPlaybackQueueCount] = song.queue_length
	nowplaying_info[MPNowPlayingInfoPropertyPlaybackQueueIndex] = song.queue_index
	nowplaying_info[MPMediaItemPropertyPersistentID] = song_to_persistent_id(song)

	nowplaying_info[MPMediaItemPropertyTitle] = song.title
	nowplaying_info[MPMediaItemPropertyArtist] = song.artist
	nowplaying_info[MPMediaItemPropertyAlbumTitle] = song.album
	nowplaying_info[MPMediaItemPropertyAlbumTrackNumber] = song.track
	nowplaying_info[MPMediaItemPropertyDiscNumber] = song.disc
	nowplaying_info[MPMediaItemPropertyGenre] = song.genre
	nowplaying_info[MPMediaItemPropertyComposer] = song.composer
	nowplaying_info[MPMediaItemPropertyPlaybackDuration] = song.duration

	# MPD can't play back music at different rates, so we just want to set it
	# to 1.0 if the song is playing. (Leave it at 0.0 if the song is paused.)
	if song.state == PlaybackState.play:
		nowplaying_info[MPNowPlayingInfoPropertyPlaybackRate] = 1.0

	if song.art:
		nowplaying_info[MPMediaItemPropertyArtwork] = ns_image_to_media_item_artwork(
			data_to_ns_image(song.art)
		)
	return nowplaying_info


def nothing_to_media_item() -> NSMutableDictionary:
	nowplaying_info = NSMutableDictionary.dictionary()
	nowplaying_info[MPNowPlayingInfoPropertyMediaType] = MPNowPlayingInfoMediaTypeNone
	nowplaying_info[MPMediaItemPropertyArtwork] = MPD_LOGO
	nowplaying_info[MPMediaItemPropertyTitle] = "MPD (stopped)"
	nowplaying_info[MPNowPlayingInfoPropertyPlaybackRate] = 0.0

	return nowplaying_info


MPD_LOGO = ns_image_to_media_item_artwork(logo_to_ns_image())


class CocoaNowPlaying:
	def __init__(self, player: Player):
		self.cmd_center = MPRemoteCommandCenter.sharedCommandCenter()
		self.info_center = MPNowPlayingInfoCenter.defaultCenter()

		cmds = (
			(self.cmd_center.togglePlayPauseCommand(), player.on_play_pause),
			(self.cmd_center.playCommand(), player.on_play),
			(self.cmd_center.pauseCommand(), player.on_pause),
			(self.cmd_center.stopCommand(), player.on_stop),
			(self.cmd_center.nextTrackCommand(), player.on_next),
			(self.cmd_center.previousTrackCommand(), player.on_prev),
		)

		for cmd, handler in cmds:
			cmd.setEnabled_(True)
			cmd.removeTarget_(None)
			cmd.addTargetWithHandler_(self._create_handler(handler))

		seekCmd = self.cmd_center.changePlaybackPositionCommand()
		seekCmd.setEnabled_(True)
		seekCmd.removeTarget_(None)
		seekCmd.addTargetWithHandler_(self._create_seek_handler(player.on_seek))

		unsupported_cmds = (
			self.cmd_center.changePlaybackRateCommand(),
			self.cmd_center.seekBackwardCommand(),
			self.cmd_center.skipBackwardCommand(),
			self.cmd_center.seekForwardCommand(),
			self.cmd_center.skipForwardCommand(),
		)
		for cmd in unsupported_cmds:
			cmd.setEnabled_(False)

		# If MPD is paused when this bridge starts up, we actually want the now
		# playing info center to see a playing -> paused transition, so we can
		# unpause with remote commands.
		self.info_center.setPlaybackState_(MPMusicPlaybackStatePlaying)

	def update(self, song: Song | None) -> None:
		if song:
			self.info_center.setNowPlayingInfo_(song_to_media_item(song))
			self.info_center.setPlaybackState_(playback_state_to_cocoa(song.state))
		else:
			self.info_center.setNowPlayingInfo_(nothing_to_media_item())
			self.info_center.setPlaybackState_(MPMusicPlaybackStateStopped)

	def _create_handler(
		self, player: Callable[[], Coroutine[None, None, PlaybackState | None]]
	) -> Callable[[MPRemoteCommandEvent], MPRemoteCommandHandlerStatus]:
		async def invoke_music_player() -> None:
			result = await player()
			if result:
				self.info_center.setPlaybackState_(playback_state_to_cocoa(result))

		def handler(event: MPRemoteCommandEvent) -> MPRemoteCommandHandlerStatus:
			run_background_task(invoke_music_player())
			return 0

		return handler

	def _create_seek_handler(
		self, player: Callable[[float], Coroutine[None, None, None]]
	) -> Callable[[MPChangePlaybackPositionCommandEvent], MPRemoteCommandHandlerStatus]:
		def handler(
			event: MPChangePlaybackPositionCommandEvent,
		) -> MPRemoteCommandHandlerStatus:
			run_background_task(player(event.positionTime()))
			return MPRemoteCommandHandlerStatusSuccess

		return handler



from __future__ import annotations

from typing import TypedDict

from ..async_tools import run_background_task
from ..cache import Cache, make_cache
from .types import CurrentSongResponse, MpdStateHandler

CACHE_TTL = 60 * 60 # seconds = 1 hour


class ArtCacheEntry(TypedDict):
	data: bytes | None


def calc_album_key(song: CurrentSongResponse) -> str:
	artist = song.get("albumartist", song.get("artist", "Unknown Artist"))
	album = song.get("album", "Unknown Album")
	return ":".join(t.replace(":", "-") for t in (artist, album))


def calc_track_key(song: CurrentSongResponse) -> str:
	return song["file"]


class MpdArtworkCache:
	mpd: MpdStateHandler
	album_cache: Cache[ArtCacheEntry]
	track_cache: Cache[ArtCacheEntry]

	def __init__(self, mpd: MpdStateHandler, cache_url: str = "memory://"):
		self.mpd = mpd
		self.album_cache = make_cache(cache_url, "album")
		self.track_cache = make_cache(cache_url, "track")

	async def get_cached_artwork(self, song: CurrentSongResponse) -> bytes | None:
		art = await self.track_cache.get(calc_track_key(song))
		if art:
			return art["data"]

		# If we don't have track artwork cached, go find some.
		run_background_task(self.cache_artwork(song))

		# Even if we don't have cached track art, we can try looking for cached album art.
		art = await self.album_cache.get(calc_album_key(song))
		if art:
			return art["data"]

		return None

	async def cache_artwork(self, song: CurrentSongResponse) -> None:
		art = ArtCacheEntry(data=await self.mpd.readpicture(song["file"]))
		try:
			await self.album_cache.add(calc_album_key(song), art, ttl=CACHE_TTL)
		except ValueError:
			pass
		await self.track_cache.set(calc_track_key(song), art, ttl=CACHE_TTL)
		await self.mpd.refresh()



from __future__ import annotations

from contextlib import suppress
from typing import Any, Optional, TypeVar
from urllib.parse import parse_qsl, urlparse

from aiocache import Cache
from aiocache.serializers import BaseSerializer, PickleSerializer

T = TypeVar("T")

HAS_ORMSGPACK = False
with suppress(ImportError):
	import ormsgpack

	HAS_ORMSGPACK = True


class OrmsgpackSerializer(BaseSerializer):
	DEFAULT_ENCODING = None

	def dumps(self, value: Any) -> bytes:
		return ormsgpack.packb(value)

	def loads(self, value: Optional[bytes]) -> Any:
		if value is None:
			return None
		return ormsgpack.unpackb(value)


def make_cache(url: str, namespace: str = "") -> Cache[T]:
	parsed_url = urlparse(url)
	backend = Cache.get_scheme_class(parsed_url.scheme)
	if backend == Cache.MEMORY:
		return Cache(backend)
	kwargs: dict[str, Any] = dict(parse_qsl(parsed_url.query))

	if parsed_url.path:
		kwargs.update(backend.parse_uri_path(parsed_url.path))

	if parsed_url.hostname:
		kwargs["endpoint"] = parsed_url.hostname

	if parsed_url.port:
		kwargs["port"] = parsed_url.port

	if parsed_url.password:
		kwargs["password"] = parsed_url.password

	namespace = ":".join(s for s in [kwargs.pop("namespace", ""), namespace] if s)

	serializer = OrmsgpackSerializer if HAS_ORMSGPACK else PickleSerializer

	return Cache(backend, serializer=serializer(), namespace=namespace, **kwargs)
